"""
Houses the :class:`_Spinner` class. This module is considered internal and should not be imported outside of the
library.
"""

from __future__ import annotations

from typing import Union

from halo import Halo


class _Spinner:
    """
    A wrapper around :class:`halo.Halo`. This class is considered internal and should not be imported outside of the
    library.
    """
    def __init__(self, spinner: Union[Halo, None]):
        self.spinner = spinner

    @staticmethod
    def spin_condition(func):
        def wrapper(spinner, *args, **kwargs):
            if spinner.spinner:
                return func(spinner, *args, **kwargs)
            else:
                return spinner

        return wrapper

    @spin_condition
    def __enter__(self):
        return self.spinner.start()

    @spin_condition
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.spinner.stop()

    @spin_condition
    def succeed(self):
        self.spinner.succeed()

    @spin_condition
    def fail(self):
        self.spinner.fail()



"""
Predefined authenticators for popular services.
"""

from loctocat.auth import Authenticator

__all__ = ["GitHubAuthenticator"]


class GitHubAuthenticator(Authenticator):
    """
    Authenticate users with GitHub.

    Parameters
    ----------
    client_id : str
        Your GitHub OAuth app's client ID.
    scopes : list[str], optional
        The scopes to request.
    """

    def __init__(self, client_id: str, scopes: list[str] = None):
        self.client_id = client_id
        self.scope = scopes

        auth_url = "https://github.com/login/device/code"
        token_url = "https://github.com/login/oauth/access_token"

        super().__init__(client_id, auth_url, token_url, scopes)



"""
Core classes for OAuth 2.0 device flow authentication.
"""

from __future__ import annotations

import time
from typing import Callable

import aiohttp
import requests
from halo import Halo
from oauthlib.oauth2 import DeviceClient
from termcolor import colored

from loctocat._spinner import _Spinner
from loctocat.exceptions import *

__all__ = ["Authenticator", "AsyncAuthenticator", "Handler", "LoctocatAuthInfo"]


class Handler:
    """
    A handler for an error code returned by an authorization server.

    Parameters
    ----------
    error: str
        The error code returned by the authorization server.
    action: Callable[[Authenticator], None]
        The function to be called when the error code is returned. The function
        must take an class:`Authenticator` object as its only parameter.
    continue_on_error: bool, optional
        Whether to continue the authentication process after the handler is
        called. Defaults to True.
    """

    def __init__(self, error: str, action: Callable[[Authenticator], None], continue_on_error: bool = True):
        self.error = error
        self.action = action
        self.continue_on_error = continue_on_error


class LoctocatAuthInfo:
    """
    A set of information returned by an authorization server.

    Parameters
    ----------
    device_code: str
        The device code.
    user_code: str
        The user code. This is the code the user must use to authorize your app.
    verificiation_uri: str
        The URL the user must visit to authorize your app.
    expires_in: int
        The number of seconds for which the device code is valid.
    interval: int
        The minimum number of seconds your app must wait between polling requests.
    """
    __slots__ = ["device_code", "user_code", "verification_uri", "expires_in", "interval"]

    def __init__(self, device_code, user_code, verificiation_uri, expires_in, interval):
        self.device_code = device_code
        self.user_code = user_code
        self.verification_uri = verificiation_uri
        self.expires_in = expires_in
        self.interval = interval


class Authenticator:
    """
    A class for authenticating with OAuth 2.0 APIs using the device flow.

    Parameters
    ----------
    client_id: str
        Your app's client ID.
    client_secret: str
        Your app's client secret.
    auth_url: str
        The URL from which your app will request device and user codes.
    token_url: str
        The URL which your app with poll for an authorization token.
    scopes: list[str], optional
        A list of scopes your app will request access to. Defaults to None.
    poll_interval: int, optional
        The number of seconds your app will wait between polling requests. Defaults to 5.
    extras: keyword arguments, optional
        Any additional parameters required by the authorization server.
    """

    def __init__(self, client_id: str, auth_url: str, token_url: str, scopes: list[str] = None, poll_interval: int = 5,
                 **extras):
        self.client_id = client_id
        self.auth_url = auth_url
        self.token_url = token_url
        self.scopes = scopes
        self.poll_interval = poll_interval

        self.auth_info: LoctocatAuthInfo = None

        self._handlers = []
        self._dc = DeviceClient(client_id=self.client_id, scopes=self.scopes, **extras)

        for handler in self._get_default_handlers():
            self.attach_handler(handler)

    def authenticate(self,
                     use_default_message: bool = True, message: str = None,
                     use_default_success_message: bool = True, success_message: str = None,
                     use_default_spinner: bool = True, spinner: Halo = None) -> str:
        """
        Authenticate with the authorization server.

        Parameters
        ----------
        use_default_message: bool, optional
            Whether to use the default message. See the message parameter. Defaults to True.
        message: str, optional
            The message to display to the user when requesting authorization. Defaults to None. If a truthy value is
            provided, it will override the default message even if use_default_message is True.
        use_default_success_message: bool, optional
            Whether to use the default success message. See the success_message parameter. Defaults to True.
        success_message: str, optional
            The message to display to the user upon successful authentication. Defaults to None. If a truthy value is
            provided, it will override the default success message even if use_default_success_message is True.
        use_default_spinner: bool, optional
            Whether to use the default spinner. See the spinner parameter. Defaults to True.
        spinner: Halo, optional
            The Halo spinner to display while waiting for the user to authorize your app. Defaults to None. If a truthy
            value is provided, it will override the default spinner even if use_default_spinner is True.

        Returns
        -------
        str
            The access token returned by the authorization server.
        """
        self.ping()

        if use_default_message and not message:
            print(f"Go to {self.auth_info.verification_uri} and enter code "
                  f"{colored(self.auth_info.user_code, attrs=['bold'])} to authenticate.")
        elif message:
            print(message)

        if use_default_spinner and not spinner:
            spinner = Halo(text="Waiting for authentication...", spinner="dots")

        with _Spinner(spinner) as sp:
            token = self.poll()

            if use_default_success_message and not success_message:
                sp.succeed("Authentication successful!")
            elif success_message:
                sp.succeed(success_message)
            else:
                sp.succeed()

            return token

    def ping(self) -> LoctocatAuthInfo:
        """
        Request device and user codes from the authorization server.

        Returns
        -------
        LoctocatAuthInfo
            A LocotocatAuthInfo object containing the information returned by the authorization server.
        """
        uri = self._dc.prepare_request_uri(self.auth_url)
        response = requests.post(uri, headers={"Accept": "application/json"}).json()

        auth_info = LoctocatAuthInfo(
            device_code=response["device_code"],
            user_code=response["user_code"],
            verificiation_uri=response.get("verification_uri") or response.get("verification_url"),
            expires_in=response["expires_in"],
            interval=response["interval"],
        )

        self.auth_info = auth_info

        return auth_info

    def poll(self) -> str:
        """
        Poll the authorization server for an access token.

        Returns
        -------
        str
            The access token returned by the authorization server.
        """
        uri = self._dc.prepare_request_uri(self.token_url, device_code=self.auth_info.device_code)

        while True:
            response = requests.post(uri, headers={"Accept": "application/json"}).json()

            if "error" in response:
                handler: Handler = next((h for h in self._handlers if h.error == response["error"]), None)
                if handler:
                    handler.action(self)
                    if handler.continue_on_error:
                        continue
                    else:
                        break
                else:
                    raise ServerError(response['error'])
            else:
                return response["access_token"]

    def attach_handler(self, handler: Handler):
        """
        Attach a handler to the Authenticator. If you attach a handler for an error for which a handler already exists,
        the existing handler will be overwritten.

        Parameters
        ----------
        handler: Handler
            The handler to attach.
        """
        try:
            self._handlers.remove(next(h for h in self._handlers if h.error == handler.error))
        except StopIteration:
            pass

        self._handlers.append(handler)

    def get_handlers(self) -> list[Handler]:
        """
        Return a list of all handlers currently attached to the Authenticator.

        Returns
        -------
        list[Handler]
            A list of all handlers currently attached to the Authenticator.
        """
        return self._handlers

    @staticmethod
    def _get_default_handlers():
        return [
            Handler("authorization_pending", lambda ctx: time.sleep(ctx.auth_info.interval)),
            Handler("slow_down", lambda ctx: ctx.poll_interval == ctx.poll_interval + 5),
        ]


class AsyncAuthenticator(Authenticator):
    """
    A class for asynchronously authenticating with OAuth 2.0 APIs using the device flow.
    """

    async def authenticate(self,
                           use_default_message: bool = True, message: str = None,
                           use_default_success_message: bool = True, success_message: str = None,
                           use_default_spinner: bool = True, spinner: Halo = None) -> str:
        """
        Authenticate with the authorization server.

        Parameters
        ----------
        use_default_message: bool, optional
            Whether to use the default message. See the message parameter. Defaults to True.
        message: str, optional
            The message to display to the user when requesting authorization. Defaults to None. If a truthy value is
            provided, it will override the default message even if use_default_message is True.
        use_default_success_message: bool, optional
            Whether to use the default success message. See the success_message parameter. Defaults to True.
        success_message: str, optional
            The message to display to the user upon succesful authentication. Defaults to None. If a truthy value is
            provided, it will override the default success message even if use_default_success_message is True.
        use_default_spinner: bool, optional
            Whether to use the default spinner. See the spinner parameter. Defaults to True.
        spinner: Halo, optional
            The Halo spinner to display while waiting for the user to authorize your app. Defaults to None. If a truthy
            value is provided, it will override the default spinner even if use_default_spinner is True.

        Returns
        -------
        str
            The access token returned by the authorization server.
        """
        await self.ping()

        if use_default_message and not message:
            print(f"Go to {self.auth_info.verification_uri} and enter code "
                  f"{colored(self.auth_info.user_code, attrs=['bold'])} to authenticate.")
        elif message:
            print(message)

        if use_default_spinner and not spinner:
            spinner = Halo(text="Waiting for authentication...", spinner="dots")

        with _Spinner(spinner) as sp:
            token = await self.poll()

            if use_default_success_message and not success_message:
                sp.succeed("Authentication successful!")
            elif success_message:
                sp.succeed(success_message)
            else:
                sp.succeed()

            return token

    async def ping(self) -> LoctocatAuthInfo:
        """
        Request device and user codes from the authorization server.

        Returns
        -------
        LoctocatAuthInfo
            A LocotocatAuthInfo object containing the information returned by the authorization server.
        """
        uri = self._dc.prepare_request_uri(self.auth_url)

        async with aiohttp.ClientSession() as session:
            async with session.post(uri, headers={"Accept": "application/json"}) as response:
                response = await response.json()

        auth_info = LoctocatAuthInfo(
            device_code=response["device_code"],
            user_code=response["user_code"],
            verificiation_uri=response.get("verification_uri") or response.get("verification_url"),
            expires_in=response["expires_in"],
            interval=response["interval"],
        )

        self.auth_info = auth_info

        return auth_info

    async def poll(self) -> str:
        """
        Poll the authorization server for an access token.

        Returns
        -------
        str
            The access token returned by the authorization server.
        """
        uri = self._dc.prepare_request_uri(self.token_url, device_code=self.auth_info.device_code)

        while True:
            async with aiohttp.ClientSession() as session:
                async with session.post(uri, headers={"Accept": "application/json"}) as response:
                    response = await response.json()

            if "error" in response:
                handler: Handler = next((h for h in self._handlers if h.error == response["error"]), None)
                if handler:
                    handler.action(self)
                else:
                    raise ServerError(f"The server returned the following error: {response['error']}")
            else:
                return response["access_token"]



"""
loctocat is the easiest way to add OAuth 2.0 device flow authentication to your Python apps.

https://loctocat.celsiusnarhwal.dev
"""

from loctocat.auth import *
from loctocat.exceptions import *
from loctocat.predefined import *

__version__ = "1.0.0"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['loctocat']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'halo>=0.0.31,<0.0.32',
 'oauthlib>=3.2.2,<4.0.0',
 'requests>=2.28.1,<3.0.0',
 'termcolor>=2.1.1,<3.0.0']

setup_kwargs = {
    'name': 'loctocat',
    'version': '1.0.2',
    'description': 'Simple yet flexible OAuth 2.0 device flow authentication for Python',
    'long_description': '# loctocat\n\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/loctocat?logo=pypi&logoColor=white&style=for-the-badge)](https://pypi.org/project/loctocat)\n[![PyPI](https://img.shields.io/pypi/v/loctocat?color=green&logo=pypi&style=for-the-badge&logoColor=white)](https://pypi.org/project/loctocat)\n[![GitHub release (latest SemVer including pre-releases)](https://img.shields.io/github/v/release/celsiusnarhwal/loctocat?color=orange&include_prereleases&label=latest%20release&logo=github&style=for-the-badge)](https://github.com/celsiusnarhwal/loctocat/releases/latest)\n[![PyPI - License](https://img.shields.io/pypi/l/loctocat?color=03cb98&style=for-the-badge)](https://github.com/celsiusnarhwal/loctocat/blob/HEAD/LICENSE.md)\n\nloctocat brings simple yet flexible OAuth 2.0 device flow authentication to Python. It has built-in asyncio support\nand even predefined authenticators for popular services. Plus, it\'s fully compliant with\n[RFC 8628](https://tools.ietf.org/html/rfc8628), making it compatible with any OAuth2-supporting service that\n(correctly) implements the standard.\n\n## Installation\n\n```bash\npip install loctocat\n```\n\n## Basic Usage\n\n### The `Authenticator` Class\n\nEvery authentication flow starts with loctocat\'s `Authenticator` class.\n\n```python\nfrom loctocat import Authenticator\n\nauthenticator = Authenticator(\n    client_id="your_client_id",\n    auth_url="https://example.com/oauth2/authorize",\n    token_url="https://example.com/oauth2/token",\n    scopes=["list", "of", "scopes"],\n)\n```\n\nIt\'s pretty simple — just instantiate the class with your client ID, authorization URL (where you\'ll get your device\nand user codes), token URL (where you\'ll poll the authorization server for an access token), and a list of any scopes\nyou need.\n\nOnce you\'ve got an `Authenticator`, getting an access token is as simple as:\n\n```python\ntoken = authenticator.authenticate()\n```\n\nWhoa. That was easy.\n\n`Authenticator.authenticate()` will, in order:\n\n1. Obtain device and user codes from the authorization server\n2. Prompt the user to visit the verficiation URL and enter the user code\n3. Poll the authorization server for an access token\n4. Return the access token as a string\n\nHere\'s an example of using `Authenticator` to authenticate with GitHub:\n\n```python\n# https://docs.github.com/en/developers/apps/building-oauth-apps/authorizing-oauth-apps#device-flow\n\nfrom loctocat import Authenticator\n\nauthenticator = Authenticator(\n    client_id="github_client_id",  # Replace this with your app\'s actual client ID, obviously.\n    auth_url="https://github.com/login/device/code",\n    token_url="https://github.com/login/oauth/access_token",\n    scopes=["repo"],  # https://docs.github.com/en/developers/apps/building-oauth-apps/scopes-for-oauth-apps\n)\n\ntoken = authenticator.authenticate()\n```\n\nLike I said, easy. Unless you\'re building an asynchronous application, in which case this doesn\'t work at all, since\n`Authenticator.authenticate()` is a blocking call. Fortunately, loctocat has you covered.\n\n### The `AsyncAuthenticator` Class\n\n`AsyncAuthenticator` is a subclass of `Authenticator` that functions exactly the same except all its methods are\nasynchronous (and therefore must be called with `await`). Here\'s an example of `AsyncAuthenticator` in action:\n\n```python\nfrom loctocat import AsyncAuthenticator\n\nauthenticator = AsyncAuthenticator(\n    client_id="your_client_id",\n    auth_url="https://example.com/oauth2/authorize",\n    token_url="https://example.com/oauth2/token",\n    scopes=["list", "of", "scopes"],\n)\n\ntoken = await authenticator.authenticate()\n```\n\nWhoa. That was easy.\n\n`AsyncAuthenticator.authenticate()` will, in or—wait, I\'m getting déjà vu.\n\n## Advanced Usage\n\nMaybe `Authenticator.authenticate()` is too simplistic for you. Maybe you\'d rather, I don\'t know,\nhandle the user-facing authentication prompt yourself, or control when loctocat starts polling for access tokens.\nFortunately, locotcat has you covered.\n\n(Keep in mind that `AsyncAuthenticator` is a subclass of `Authenticator` and inherits all of its methods and\nattributes.)\n\n### `Authenticator.ping()`\n\n`Authenticator.ping()` requests device and user codes from the authorization server, returning a `LoctocatAuthInfo`\nobject that looks like this:\n\n```py\nclass LoctocatAuthInfo:\n    device_code: str\n    user_code: str\n    verification_uri: str\n    expires_in: int\n    interval: int\n```\n\nPretty self-explanatory. You can do whatever you want with this information (aside from change it\n— `LoctocatAuthInfo`\'s attributes are read-only). For example, you could prompt the user with some custom text\ncontaining the user code and verification URI:\n\n```python\nauth_info = authenticator.ping()\n\nprint(f"Check it out, yo! This is some epic text telling YOU to go {auth_info.verification_uri} and enter {auth_info.user_code}! Swag!")\n```\n\n### `Authenticator.poll()`\n\n`Authenticator.poll()` polls the authorization server for an access token, returning it as a string. You **don\'t**\nneed to pass the `LoctocatAuthInfo` object returned by `Authenticator.ping()` to `Authenticator.poll()` — the\nauthorization info is automatically remembered by `Authenticator`. All you have to do is call the method:\n\n```python\n# Authenticator.ping() must have been called on the Authenticator object already or this will not work.\n\ntoken = authenticator.poll()\n```\n\nWhoa. That was easy.\n\n(Fun fact: `Authenticator.authenticate()` is just a wrapper around `ping()` and `poll()`.)\n\n## Pro Usage\n\nMaybe [Advanced Usage](#advanced-usage) isn\'t advanced enough for you. Maybe you\'re working with an authorization\nserver that requires parameters beyond those defined by `Authenticator`. Maybe you want to customize the prompts and\nmessages displayed by `Authenticator.authenticate()` without having to use `Authenticator.ping()` and\n`Authenticator.poll()`. Maybe loctocat has a predefined authenticator for a service you like, and you want to use it.\nUnfortunately, loctocat doesn\'t have you covered.\n\n...\n\nOkay, loctocat actually does have you covered, but that stuff is the domain of loctocat\'s unfinished documentation\nsite. Emphasis on unfinished. It\'s not finished yet.\n\nFortunately, loctocat has you covered. loctocat\'s a pretty small library and it\'s public modules and classes are all\nproperly documented in the source code, so you\'re welcome to learn by example(?) and take a look around.\n\nOr you could wait until I finish the documentation site. I\'m not your mother.\n\n## FAQ\n\nMaybe you have questions about loctocat that haven\'t been answered by the rest of this README. Maybe you just want to see me talk to\nmyself for like, two paragraphs. Fortunately, loctocat has you covered.\n\n### Q: loctowhat now\n\nA: Lock + [Octocat](https://octodex.github.com). loctocat was born out of my need for a Python library that implemented\nOAuth 2.0 device flow authentication for GitHub.\n\n### Q: I can just do this with requests-oauthlib or [INSERT OAUTH LIBRARY HERE] and those are like wayyyy more popular and tested so ummmm owned much?\n\nA: Ratio.\n\n### Q: loctocat isn\'t working with [INSERT SERVICE HERE] and I\'m FRUSTRATED AAAAGGGGGGHHHHH\n\nA: loctocat is compliant with the OAuth 2.0 Device Authorization Grant standard so it\'s probably the service\'s fault.\nMake sure the service actually does support the device flow and is generally compliant with RFC 8628. If you\'re sure\nloctocat is the problem, [open an issue](https://github.com/celsiusnarhwal/loctocat/issues/new).\n\n### Q: oh my god thank you I\'ve been looking for a library like this forever you have no idea\n\nA: You\'re very welcome. 🙂\n\n## License\n\nIn an age where developers must take great caution not to tread on the intellectual property of others, you must be\nhoping that a such an incredible library is made available under a permissive license. Fortunately,\nloctocat has you covered.\n\nloctocat is licensed under the [MIT License](https://github.com/celsiusnarhwal/loctocat/blob/HEAD/LICENSE.md).\n',
    'author': 'celsius narhwal',
    'author_email': 'hello@celsiusnarhwal.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/celsiusnarhwal/loctocat',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
